<?php
require_once 'app.class.php';
require_once 'users.class.php';
require_once 'database.class.php';
require_once 'errors.class.php';
require_once 'mailer.class.php';
require_once 'paginator.class.php';

class Transactions
{
    private $conn;
    private const MONTHS = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];

    public function __construct()
    {
        $this->conn = Database::connect();
    }

    private function isValidPlan(int $planId): bool
    {
        $valid = false;
        $plans = App::INVESTMENT_PLANS;
        foreach ($plans as $plan) {
            if ($plan['id'] === $planId) {
                $valid = true;
                break;
            }
        }
        return $valid;
    }

    private function isValidCurrency(int $currencyId): bool
    {
        $valid = false;
        $currencies = App::CURRENCIES;
        foreach ($currencies as $currency) {
            if ($currency['id'] === $currencyId) {
                $valid = true;
                break;
            }
        }
        return $valid;
    }

    public function createInvestment(int $planId, float $amount, int $currencyId, int $userID): array
    {
        $response = ['success' => null, 'error' => null];

        if ($this->isValidPlan($planId)) {
            $planIds = array_map(function ($plan) {
                return $plan['id'];
            }, App::INVESTMENT_PLANS);
            $planIndex = array_search($planId, $planIds);
            $plan = App::INVESTMENT_PLANS[$planIndex];

            if ($amount >= $plan['minimumDeposit']) {
                if ($amount <= $plan['maximumDeposit']) {
                    $sql = "SELECT id FROM deposits WHERE investment_plan_id = ? AND `user_id` = ?";
                    $prepStmt = $this->conn->prepare($sql);
                    $prepStmt->bind_param('ii', $planId, $userID);
                    $prepStmt->execute();
                    $result = $prepStmt->get_result();
                    $numOfInvestmentMade = $result->num_rows;
                    if($plan['maximimuInvestmentCount'] !== null){                    if ($numOfInvestmentMade >= $plan['maximimuInvestmentCount']) {
                        $response['error'] = 'Maximum investment number reached for this plan. Try other investment plan';
                    }
                    }
                } else {
                    $response['error'] = 'Amount is higher than maximum deposit';
                }
            } else {
                $response['error'] = 'Amount is lower than minimum deposit';
            }
            if ($response['error'] === null) {

                if ($this->isValidCurrency($currencyId)) {
                    $sql = "INSERT INTO deposits
                (amount, investment_plan_id, `user_id`, is_payment_made, currency_id, is_payment_confirmed, date_generated)
                VALUES(?, ?, ?, 0, ?, 0, CURRENT_TIMESTAMP);
                ";
                    $prepStmt = $this->conn->prepare($sql);
                    $prepStmt->bind_param('diii', $amount, $planId, $userID, $currencyId);
                    if ($prepStmt->execute()) {
                        $response['success'] = true;
                    } else {
                        $response['error'] = Errors::transactionError('cannotCompleteTransaction');
                    }
                } else {
                    $response['error'] = Errors::transactionError('invalidCurrency');
                }
            }
        } else {
            $response['error'] = Errors::transactionError('invalidPlan');
        }
        return $response;
    }

    public function getInvestments(int $page, int $limit, int $userId = null, $email = ''): array
    {
        $response = ['data' => []];
        if (trim($email) !== "") {
            $user = new Users();
            $userId = $user->getUserId(trim($email));
        }
        $sql = "SELECT deposits.id, deposits.amount, deposits.investment_plan_id, deposits.user_id, deposits.is_payment_made, deposits.currency_id, deposits.is_payment_confirmed, deposits.date_generated, deposits.date_confirmed, deposits.profits , users.first_name, users.last_name,
        users.currencies_accounts, users.email FROM deposits JOIN users ON deposits.user_id = users.id";
        $sql .= $userId !== null ? " WHERE deposits.user_id = $userId" : "";
        $sql .= " ORDER BY deposits.id DESC;";
        $paginator = new Paginator(Database::connect(), $limit);
        $results = $paginator->paginate($page, $sql);

        foreach ($results['data'] as $r) {
            $planIds = array_map(function ($plan) {
                return $plan['id'];
            }, App::INVESTMENT_PLANS);

            $planIndex = array_search($r['investment_plan_id'], $planIds);

            $currencyIds = array_map(function ($currency) {
                return $currency['id'];
            }, App::CURRENCIES);

            $currencyIndex = array_search($r['currency_id'], $currencyIds);
            array_push($response['data'], [
                'id' => $r['id'],
                'plan' => App::INVESTMENT_PLANS[$planIndex],
                'currency' => App::CURRENCIES[$currencyIndex],
                'amount' => $r['amount'],
                'isPaymentMade' => $r['is_payment_made'],
                'isPaymentConfirmed' => $r['is_payment_confirmed'],
                'dateGenerated' => $r['date_generated'],
                'dateConfirmed' => $r['date_confirmed'],
                'profits' => $r['profits'] !== null ? json_decode($r['profits'], true) : [],
                'user' => [
                    'id' => $r['user_id'],
                    'firstName' => $r['first_name'],
                    'lastName' => $r['last_name'],
                    'email' => $r['email'],
                    'currencies' => json_decode($r['currencies_accounts'], true),
                ],
            ]);
        }

        $response['info'] = $results['info'];
        return $response;
    }

    public function confirmInvestmentPayment(int $id): array
    {
        $response = ['success' => false];
        $sql = "UPDATE deposits SET is_payment_made = 1 WHERE id = ?";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        if($prepStmt->execute()){
            $sql = "SELECT deposits.user_id, deposits.amount, users.first_name, users.email FROM deposits JOIN users ON deposits.user_id = users.id WHERE deposits.id = ? GROUP BY deposits.id";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('i', $id);
            $prepStmt->execute();
            $result = $prepStmt->get_result();
            if($result->num_rows === 1){
                $row = $result->fetch_assoc();
                $mailer = new Mailer();
                $mailer->sendConfirmPaymentMail($row['email'], (float) $row['amount'], $row['first_name']);
            }
            
           $response['success'] = true;
        }
        return $response;
    }

    public function approveInvestmentPayment(int $id): array
    {
        $response = ['success' => false];
        $sql = "UPDATE deposits SET is_payment_confirmed = 1, date_confirmed = CURRENT_TIMESTAMP WHERE id = ?";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        if($prepStmt->execute()){
            $sql = "SELECT deposits.user_id, deposits.amount, users.first_name, users.email FROM deposits JOIN users ON deposits.user_id = users.id WHERE deposits.id = ? GROUP BY deposits.id";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('i', $id);
            $prepStmt->execute();
            $result = $prepStmt->get_result();
            if($result->num_rows === 1){
                $row = $result->fetch_assoc();
                $mailer = new Mailer();
                $mailer->sendInvestmentAprrovalMail($row['email'], (float) $row['amount'], $row['first_name']);
            }
            $response['success'] = true;
        }
        return $response;
    }

    public function deleteInvestmentPayment(int $id): array
    {
        $response = ['success' => false];
        $sql = "DELETE FROM deposits WHERE id = ?";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        $response['success'] = $prepStmt->execute();
        return $response;
    }

    public function addProfitToInvestment(int $id, float $amount): array
    {
        $response = ['success' => false];

        $sql = "SELECT id, profits FROM deposits WHERE id = ?";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        if ($result->num_rows === 1) {
            $row = $result->fetch_assoc();
            $profits = $row['profits'] !== null ? json_decode($row['profits'], true) : [];
            array_push($profits, [
                'date' => date('Y-m-d'),
                'amount' => $amount,
            ]);
            $profits = json_encode($profits);

            $sql = "UPDATE deposits SET profits = ? WHERE id = ?";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('si', $profits, $id);
            if($prepStmt->execute()){
                $sql = "SELECT deposits.user_id, deposits.amount, users.first_name, users.email FROM deposits JOIN users ON deposits.user_id = users.id WHERE deposits.id = ? GROUP BY deposits.id";
                $prepStmt = $this->conn->prepare($sql);
                $prepStmt->bind_param('i', $id);
                $prepStmt->execute();
                $result = $prepStmt->get_result();
                if($result->num_rows === 1){
                    $row = $result->fetch_assoc();
                    $mailer = new Mailer();
                    $mailer->sendProfitMail($row['email'], $amount, (float) $row['amount'], $row['first_name']);
                }
            
                $response['success'] = true;
            }
        } else {
            $response['error'] = Errors::transactionError('unknownInvestment');
        }

        return $response;
    }

    public function getAccountBalanceSummary(int $userID): array
    {
        $response = [
            'totalDeposits' => 0,
            'totalProfits' => 0,
            'totalWithdrawals' => 0,
            'availableBalance' => 0,
        ];
        $sql = "SELECT amount, profits FROM deposits WHERE `user_id` = ? AND is_payment_confirmed = 1;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $userID);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        if ($result->num_rows > 0) {
            $deposits = $result->fetch_all(MYSQLI_ASSOC);
            foreach ($deposits as $d) {
                $response['totalDeposits'] += $d['amount'];
                $profits = $d['profits'] !== null ? json_decode($d['profits'], true) : [];

                foreach ($profits as $p) {
                    $response['totalProfits'] += $p['amount'];
                }
            }
        }

        //get withdrawals

        $sql = "SELECT amount FROM withdrawals WHERE `user_id` = ? AND date_approved IS NOT NULL";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $userID);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        if ($result->num_rows > 0) {
            $withdrawals = $result->fetch_all(MYSQLI_ASSOC);
            foreach ($withdrawals as $w) {
                $response['totalWithdrawals'] += $w['amount'];
            }
        }

        $balance = (($response['totalDeposits'] + $response['totalProfits']) - $response['totalWithdrawals']);
        $response['availableBalance'] = $balance >= 0 ? $balance : 0;

        return $response;
    }

    public function requestWithdrawal(float $amount, int $currencyId, int $userId): array
    {
        $response = ['success' => false];
        $account = $this->getAccountBalanceSummary($userId);

        if ($account['availableBalance'] >= $amount) {
            $sql = "INSERT INTO withdrawals
            (amount, `user_id`, currency_id, date_requested)
            VALUES(?, ?, ?, CURRENT_TIMESTAMP);
            ";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('dii', $amount, $userId, $currencyId);
            if($prepStmt->execute()){
                $sql = "SELECT first_name, email FROM users WHERE id=?";
                $prepStmt = $this->conn->prepare($sql);
                $prepStmt->bind_param('i', $userId);
                $prepStmt->execute();
                $result = $prepStmt->get_result();
                if($result->num_rows === 1){
                    $row = $result->fetch_assoc();
                    $mailer = new Mailer();
                    $mailer->sendWithdrawalNotificationMail($row['email'], $amount, $row['first_name']);
                }
            
                $response['success'] = true;
            }
        } else {
            $response['error'] = Errors::transactionError('insufficentFund');
        }
        return $response;
    }

    public function approveWithdrawalRequest(int $id): array
    {
        $response = ['success' => false];
        $sql = "UPDATE withdrawals SET date_approved = CURRENT_TIMESTAMP
        WHERE id=?;
        ";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        if($prepStmt->execute()){
            $sql = "SELECT withdrawals.user_id, withdrawals.amount, users.first_name, users.email FROM withdrawals JOIN users ON withdrawals.user_id = users.id WHERE withdrawals.id = ? GROUP BY withdrawals.id";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('i', $id);
            $prepStmt->execute();
            $result = $prepStmt->get_result();
            if($result->num_rows === 1){
                $row = $result->fetch_assoc();
                $mailer = new Mailer();
                $mailer->sendWithdrawalAprrovalMail($row['email'], (float) $row['amount'], $row['first_name']);
            }
            
            $response['success'] = true;
        }
        return $response;
    }

    public function deleteWithdrawalRequest(int $id): array
    {
        $response = ['success' => false];
        $sql = "DELETE FROM withdrawals WHERE id=?;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        $response['success'] = $prepStmt->execute();
        return $response;
    }

    public function getWithdrawals(int $page, int $limit, int $userId = null,
         string $email = ''): array
    {
        $response = ['data' => []];
        if (trim($email) !== "") {
            $user = new Users();
            $userId = $user->getUserId(trim($email));
        }
        

        $sql = "SELECT
        withdrawals.id, withdrawals.amount, withdrawals.user_id,
        withdrawals.currency_id, withdrawals.date_requested,
        withdrawals.date_approved, users.first_name, users.last_name,
        users.currencies_accounts, users.email, users.username
        FROM withdrawals JOIN users ON withdrawals.user_id = users.id";

        $sql .= $userId !== null ? " WHERE user_id = $userId" : "";
        $sql .= " GROUP BY withdrawals.id   ORDER BY id DESC;";
        $paginator = new Paginator(Database::connect(), $limit);
        $results = $paginator->paginate($page, $sql);
        

        foreach ($results['data'] as $r) {

            $currencyIds = array_map(function ($currency) {
                return $currency['id'];
            }, App::CURRENCIES);

            $currencyIndex = array_search($r['currency_id'], $currencyIds);
            array_push($response['data'], [
                'id' => $r['id'],
                'amount' => $r['amount'],
                'currency' => App::CURRENCIES[$currencyIndex],
                'dateRequested' => $r['date_requested'],
                'dateApproved' => $r['date_approved'],
                'user' => [
                    'id' => $r['user_id'],
                    'firstName' => $r['first_name'],
                    'lastName' => $r['last_name'],
                    'email' => $r['email'],
                    'username' => $r['username'],
                    'currencies' => json_decode($r['currencies_accounts'], true),
                ],
                'balance' => $this->getAccountBalanceSummary((int) $r['user_id'])
            ]);
        }

        $response['info'] = $results['info'];
        return $response;
    }

    public function investmentFullDetail(int $id): array
    {
        $response = [];
        $sql = "SELECT deposits.id, deposits.amount, deposits.investment_plan_id,
        deposits.is_payment_made, deposits.currency_id,
        deposits.is_payment_confirmed,
        deposits.date_generated,
        deposits.date_confirmed,
        deposits.profits, users.id as user_id, users.first_name, users.last_name, users.email
        FROM deposits JOIn users ON deposits.user_id = users.id WHERE deposits.id = ?";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $id);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        if ($result->num_rows === 1) {
            $r = $result->fetch_assoc();

            $planIds = array_map(function ($plan) {
                return $plan['id'];
            }, App::INVESTMENT_PLANS);

            $planIndex = array_search($r['investment_plan_id'], $planIds);

            $currencyIds = array_map(function ($currency) {
                return $currency['id'];
            }, App::CURRENCIES);
            $currencyIndex = array_search($r['currency_id'], $currencyIds);
            $response = [
                'investment' => [
                    'id' => $r['id'],
                    'plan' => App::INVESTMENT_PLANS[$planIndex],
                    'currency' => App::CURRENCIES[$currencyIndex],
                    'amount' => $r['amount'],
                    'isPaymentMade' => $r['is_payment_made'],
                    'isPaymentConfirmed' => $r['is_payment_confirmed'],
                    'dateGenerated' => $r['date_generated'],
                    'dateConfirmed' => $r['date_confirmed'],
                    'profits' => $r['profits'] !== null ? json_decode($r['profits'], true) : [],

                ],
                'user' => [
                    'id' => $r['user_id'],
                    'firstName' => $r['first_name'],
                    'lastName' => $r['last_name'],
                    'email' => $r['email'],
                ],
            ];
        } else {
            $response['error'] = Errors::transactionError('unknownInvestment');
        }
        return $response;
    }

    public function addProfitToInvestmentAutomatically(int $userId, string $email)
    {
        if (App::SETTINGS['autoProfit']) {
            $sql = "SELECT id, amount, investment_plan_id, profits, date_confirmed
            FROM deposits
            WHERE `user_id` = ? AND is_payment_confirmed = 1
            ORDER BY id DESC;";
            $prepStmt = $this->conn->prepare($sql);
            $prepStmt->bind_param('i', $userId);
            $prepStmt->execute();
            $result = $prepStmt->get_result();
            $deposits = $result->fetch_all(MYSQLI_ASSOC);
            foreach ($deposits as $deposit) {
                $profits = $deposit['profits'] !== null ? json_decode($deposit['profits'], true) : [];
                $lastProfitCount = count($profits);
                $planIds = array_map(function ($plan) {
                    return $plan['id'];
                }, App::INVESTMENT_PLANS);

                $planIndex = array_search($deposit['investment_plan_id'], $planIds);
                $plan = App::INVESTMENT_PLANS[$planIndex];
                if (count($profits) === 0) {
                    $hoursDurationBeforeProfit = $plan['durationInHours'];
                    $hoursSincePayment = strtotime($deposit['date_confirmed']) / (60 * 60);

                    if ($hoursSincePayment >= $hoursDurationBeforeProfit) {
                        $percentageProfit = $plan['percentIncrease'];
                        $amount = $deposit['amount'];
                        $profit = ($amount * $percentageProfit) / 100;
                        $dateOfProfit = date('Y-m-d', strtotime(strtotime($deposit['date_confirmed']) + ($hoursDurationBeforeProfit * 60 * 60)));
                        array_push($profits, [
                            'date' => $dateOfProfit,
                            'amount' => $profit,
                        ]);
                    }
                } else {
                    
                    if ($plan['maximumTurnOutCount'] === null) {
                        //  if infinite turnout count
                        while (true) {
                            $lastProfit = $profits[count($profits) - 1];
                            $hoursDurationBeforeProfit = $plan['durationInHours'];
                            $hoursSinceLastProfit = strtotime($lastProfit['date']) / (60 * 60);

                            if ($hoursSinceLastProfit >= $hoursDurationBeforeProfit) {
                                $percentageProfit = $plan['percentIncrease'];
                                $amount = $deposit['amount'];
                                $profit = ($amount * $percentageProfit) / 100;
                                $dateOfProfit = date('Y-m-d', strtotime(strtotime($lastProfit['date']) + ($hoursDurationBeforeProfit * 60 * 60)));
                                array_push($profits, [
                                    'date' => $dateOfProfit,
                                    'amount' => $profit,
                                ]);
                            } else {
                                break;
                            }
                        }

                    } else {
                        // if turnout is limited
                        $leftLimit = $plan['maximumTurnOutCount'] - count($profits);
                        for ($c = 0; $c < $leftLimit; $c++) {
                            $lastProfit = $profits[count($profits) - 1];
                            $hoursDurationBeforeProfit = $plan['durationInHours'];
                            $hoursSinceLastProfit = strtotime($lastProfit['date']) / (60 * 60);

                            if ($hoursSinceLastProfit >= $hoursDurationBeforeProfit) {
                                $percentageProfit = $plan['percentIncrease'];
                                $amount = $deposit['amount'];
                                $profit = ($amount * $percentageProfit) / 100;
                                $dateOfProfit = date('Y-m-d', strtotime(strtotime($lastProfit['date']) + ($hoursDurationBeforeProfit * 60 * 60)));
                                array_push($profits, [
                                    'date' => $dateOfProfit,
                                    'amount' => $profit,
                                ]);
                            }
                        }

                    }
                }

                if ($lastProfitCount < count($profits)) {
                    $sql = "UPDATE deposits SET profits = ? WHERE id = ?";
                    $d = json_encode($profits);
                    $id = $deposit['id'];
                    $prepStmt = $this->conn->prepare($sql);
                    $prepStmt->bind_param('si', $d, $id);
                    if ($prepStmt->execute()) {
                        $mailer = new Mailer();
                       // $mailer->sendProfitMail($email,);
                    }
                }
            }

        }
    }

    public function investmentChartForYear(int $userId): array
    {
        $response = [];

        $sql = "SELECT DISTINCT investment_plan_id FROM deposits WHERE user_id =? AND is_payment_confirmed = 1;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->bind_param('i', $userId);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        $rows = $result->fetch_all(MYSQLI_ASSOC);
        $planIds = array_map(function ($plan) {
            return $plan['id'];
        }, App::INVESTMENT_PLANS);
        foreach ($rows as $row) {
            $planId = $row['investment_plan_id'];
            $plan = [
                'plan' => null,
                'investments' => [],
            ];
            $index = array_search($planId, $planIds);

            if ($index >= 0) {
                $plan['plan'] = APP::INVESTMENT_PLANS[$index];

                $sql = "SELECT
                deposits.id, deposits.amount, deposits.investment_plan_id,
                deposits.is_payment_made, deposits.currency_id,
                deposits.is_payment_confirmed,
                deposits.date_generated,
                deposits.date_confirmed,
                deposits.profits
                FROM deposits
                WHERE  deposits.investment_plan_id =?
                AND deposits.is_payment_confirmed = 1
                AND  deposits.user_id = ?
                ORDER BY deposits.id
                DESC LIMIT 5;";
                $prepStmt = $this->conn->prepare($sql);
                $prepStmt->bind_param('ii', $planId, $userId);
                $prepStmt->execute();
                $result = $prepStmt->get_result();
                $data = $result->fetch_all(MYSQLI_ASSOC);

                foreach ($data as $r) {
                    $planIds = array_map(function ($plan) {
                        return $plan['id'];
                    }, App::INVESTMENT_PLANS);

                    $planIndex = array_search($r['investment_plan_id'], $planIds);

                    $currencyIds = array_map(function ($currency) {
                        return $currency['id'];
                    }, App::CURRENCIES);
                    $currencyIndex = array_search($r['currency_id'], $currencyIds);
                    array_push($plan['investments'], [
                        'id' => $r['id'],
                        'plan' => App::INVESTMENT_PLANS[$planIndex],
                        'currency' => App::CURRENCIES[$currencyIndex],
                        'amount' => $r['amount'],
                        'isPaymentMade' => $r['is_payment_made'],
                        'isPaymentConfirmed' => $r['is_payment_confirmed'],
                        'dateGenerated' => $r['date_generated'],
                        'dateConfirmed' => $r['date_confirmed'],
                        'profits' => $r['profits'] !== null ? json_decode($r['profits'], true) : [],

                    ]);
                }

            }

            array_push($response, $plan);
        }

        return $response;
    }

    public function transactionStatistics(): array
    {
        $response = [
            'totalDepositCount' => 0,
            'totalWithdrawCount' => 0,
            'sumDeposits' => 0,
        ];

        $sql = "SELECT id FROM deposits WHERE is_payment_confirmed = 1;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        $response['totalDepositCount'] = $result->num_rows;

        $sql = "SELECT SUM(amount) as total FROM deposits WHERE is_payment_confirmed = 1;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        $row = $result->fetch_assoc();
        $response['sumDeposits'] = $row['total'];

        $sql = "SELECT id FROM withdrawals WHERE date_approved IS NOT NULL;";
        $prepStmt = $this->conn->prepare($sql);
        $prepStmt->execute();
        $result = $prepStmt->get_result();
        $response['totalWithdrawCount'] = $result->num_rows;

        return $response;
    }

}
